#include "MarkerGroup.h"
#include "Error.h"
#include "StringArray.h"
#include <iostream>

GroupInfo::GroupInfo()
{
 start=0;
 end = 0;
 count=0;
}

GroupInfo::~GroupInfo(){}

MarkerGroup::MarkerGroup()
{
 min_freq = 0.0;
 max_freq = 1.0;
 count = 0;
}

MarkerGroup::~MarkerGroup(){}

void MarkerGroup::SetMinMaxFreq(double min, double max)
{
 min_freq = min;
 max_freq = max;
}

void MarkerGroup::LoadMarkerGroupInfo(std::string marker_info_file)
{
 FILE *fh = fopen(marker_info_file.c_str(), "r");
 if(fh==NULL) error("Open groupfile %s failed!\n", marker_info_file.c_str() );

 std::string SET("GROUP");
 std::string CHR("CHR");
 std::string POS("POS");
 std::string WEIGHT("WEIGHT");
 std::string FREQ("FREQ");

 has_weight = false;
 has_freq = false;

 String buffer;
 StringArray tokens;
 
 int ln = 0;
 while(!feof(fh) )
 {
  buffer.ReadLine(fh);
  if(buffer.Length()==0) continue;

  tokens.ReplaceTokens(buffer);
  if(tokens.Length()==0) continue;
  if(tokens.Length()<3) error("Too few columns in the marker info file!\n");

  ln++;
  if(ln==1 && tokens[0][0]!='#')
   {
    index["CHR"] = 0;
    index["POS"] = 1;
    index["GROUP"] = 2;
    fprintf(stderr, "WARNING: No header in the group file: the first 3 columns were taken as CHR POS GROUP and others were ignored!\n");
   }

  if(ln==1 && tokens[0][0]=='#')
  {
   tokens[0] = tokens[0].SubStr(1, tokens[0].Length()-1);

   for(int i=0; i<tokens.Length(); i++)
    index[std::string(tokens[i].ToUpper().c_str())] = i;

  if(index.find(WEIGHT)!=index.end()) has_weight = true;
  if(index.find(FREQ)!=index.end()) has_freq = true;

  if(index.find(SET)==index.end() || index.find(CHR)==index.end() || index.find(POS)==index.end() )
	error("CHR or POS or GROUP is missing in the header!\n");
   continue;
  }

  std::string group_name = std::string(tokens[index[SET]].c_str() );
  std::string chr_name = std::string(tokens[index[CHR]].c_str() );

  if(chr_name.compare(chrY)==0 || chr_name.compare(chrMT)==0) {
	 fprintf(stderr, "WARNING: chrY and MT variants are not handled in the current implementation and ignored... %s:%d\n", chr_name.c_str(), tokens[index[POS]].AsInteger());
	 continue; // ignore Y and MT chromsomes for now. May implement later.
  }

  if(index.find(FREQ)!=index.end())
  {
    double freq = tokens[index[FREQ]].AsDouble();
    if(freq > 0.5) freq = 1-freq;
    if(freq<min_freq || freq>max_freq) continue;
  }

  if(groups.find(group_name)==groups.end() ) group_names.push_back(group_name);
  groups[group_name].chr.push_back(chr_name);
  groups[group_name].chr_map[chr_name]++;
  groups[group_name].pos.push_back(tokens[index[POS]].AsInteger());
  if(groups[group_name].pos[groups[group_name].pos.size()-1]<=0) error("POS is invalid: %s\n", buffer.c_str());
  groups[group_name].range.push_back(chr_name+":"+std::string(tokens[index[POS]].c_str())+"-"+std::string(tokens[index[POS]].c_str()));
  if(has_weight) groups[group_name].weight.push_back(tokens[index[WEIGHT]].AsDouble());
  if(has_freq)   groups[group_name].freq.push_back(tokens[index[FREQ]].AsDouble());

  groups[group_name].count++;
 }

 CheckChromCompatability();

 for(int i=0; i<group_names.size(); i++)
 {
  groups[group_names[i]].start = groups[group_names[i]].pos[0];
  groups[group_names[i]].end = groups[group_names[i]].pos[groups[group_names[i]].count-1];

  groups[group_names[i]].weight_one.resize(groups[group_names[i]].count);
  groups[group_names[i]].weight_one.assign(groups[group_names[i]].weight_one.size(), 1.0);

  int cnt=0;
  groups[group_names[i]].chr_uniq.resize(0);
  groups[group_names[i]].chr_str = "";
  for (std::map<std::string,int>::iterator it=groups[group_names[i]].chr_map.begin(); it!=groups[group_names[i]].chr_map.end(); ++it)
  {
   cnt++;
   groups[group_names[i]].chr_uniq.push_back(it->first);
   groups[group_names[i]].chr_str += it->first;
   if(cnt<groups[group_names[i]].chr_map.size())
    groups[group_names[i]].chr_str += ";";
  }
 }
  count = group_names.size();
}

// check if a group has mixed autosomes and sex or MT chrmosomes
void MarkerGroup::CheckChromCompatability()
{

 for(int i=0; i<group_names.size(); i++)
 {
  int nChrX=0; int nChrY=0; int nChrMT=0; int nAuto=0;
  for(int j=0; j<groups[group_names[i]].count; j++)
  {
   if(groups[group_names[i]].chr[j].compare(chrX)==0) nChrX = 1;
   else if(groups[group_names[i]].chr[j].compare(chrY)==0) nChrY = 1;
   else if(groups[group_names[i]].chr[j].compare(chrMT)==0) nChrMT = 1;
   else nAuto = 1;
  }
  if(nChrX + nChrY + nChrMT + nAuto > 1) error("Group %s has mixed autosomes and non-authsomes and cannot be handled!\n", group_names[i].c_str());

  if(nChrX) groups[group_names[i]].isChrX = true; else groups[group_names[i]].isChrX = false;
  if(nChrY) groups[group_names[i]].isChrY = true; else groups[group_names[i]].isChrY = false;
  if(nChrMT) groups[group_names[i]].isChrMT = true; else groups[group_names[i]].isChrMT = false;
 }
}

void MarkerGroup::PrintGroupInfo()
{
 for(int i=0; i<group_names.size(); i++)
 {
  cout<<group_names[i]<<" : "<<endl;
  for(int j=0; j<groups[group_names[i]].count; j++)
   cout<<groups[group_names[i]].chr[j]<< " " <<groups[group_names[i]].pos[j]<< " " << groups[group_names[i]].weight[j]<<endl;
 }
}
